@@ -12,6 +12,7 @@ import colorsys
 import logging
 import math
 import sys
+import warnings
 
 
 def get_transfer_function_phi(alpha, gamma):
@@ -161,90 +162,138 @@ class HLG(object):
 	"""
 
 	def __init__(self, black_cdm2=0.0, white_cdm2=1000.0, system_gamma=1.2,
-				 rgb_space="Rec. 2020"):
+				 ambient_cdm2=5, rgb_space="Rec. 2020"):
 		self.black_cdm2 = black_cdm2
 		self.white_cdm2 = white_cdm2
 		self.rgb_space = get_rgb_space(rgb_space)
 		self.system_gamma = system_gamma
+		self.ambient_cdm2 = ambient_cdm2
+
+	@property
+	def gamma(self):
+		""" System gamma for nominal peak luminance and ambient """
+		# Adjust system gamma for peak luminance != 1000 cd/m2
+		gamma = self.system_gamma + 0.42 * math.log10(self.white_cdm2 / 1000.0)
+		if self.ambient_cdm2 > 0:
+			# Adjust system gamma for ambient surround != 5 cd/m2 (BT.2390-3)
+			gamma -= 0.076 * math.log10(self.ambient_cdm2 / 5.0)
+		return gamma
 		
-	def oetf(self, v, inverse=False, normalize=1, apply_system_gamma=True,
-			 apply_black_offset=True):
+	def oetf(self, v, inverse=False):
 		"""
 		Hybrid Log Gamma (HLG) OETF
 		
-		"""
-		if normalize == 1:
-			# Normalized to 0..1 range
-			c = 1.00429347
-			d = 1.0
-			e = 3.0
-			k = 1.0 / 12
-			r = 1.0
-		elif normalize == 12:
-			# Normalized to 0..12 range
-			c = 0.55991073
-			d = 4.0
-			e = 1.0
-			k = 1.0
-			r = 0.5
-		else:
-			raise ValueError("Cannot normalize to %s" % normalize)
-		a = 0.17883277
-		b = 0.02372241 * normalize
-		if inverse:
-			# Inverse OETF, RGB -> XYZ
-			E_ = v
-			if 0 <= E_ <= 0.5:
-				E = d * E_ ** 2 / e
-			else:
-				E = math.exp((E_ - c) / a) + b
-			v = E
-			if apply_system_gamma:
-				gamma = self.system_gamma + 0.42 * math.log10(self.white_cdm2 / 1000.0)
-				v = v ** gamma
-			if apply_black_offset:
-				LB = self.black_cdm2 / self.white_cdm2
-				alpha = (1.0 - LB)
-				v = alpha * v + LB
-		else:
-			# OETF, XYZ -> RGB
-			if apply_system_gamma:
-				gamma = self.system_gamma + 0.42 * math.log10(self.white_cdm2 / 1000.0)
-				v = (v / normalize) ** (1 / gamma) * normalize
-			E = v
-			if 0 <= E <= k:
-				E_ = r * math.sqrt(e * E)
-			else:
-				E_ = a * math.log(E - b) + c
-			v = E_
-		return min(v, normalize)
+		Relative scene linear light to non-linear HLG signal, or inverse
 
-	def eotf(self, R, G, B):
-		"""
-		Hybrid Log Gamma (HLG) EOTF
-
-		Return luminance of displayed components (in cd/m2) or XYZ
-		(normalized to 0..1) if rgb_space is given
+		Input domain 0..1
+		Output range 0..1
 		
 		"""
-		RGB_S = tuple(self.oetf(v, True, 1, False, False) for v in (R, G, B))
-		gamma = self.system_gamma + 0.42 * math.log10(self.white_cdm2 / 1000.0)
-		if self.rgb_space:
-			X, Y, Z = self.rgb_space[-1] * RGB_S
-			if Y:
-				# Apply system gamma
-				Yy = Y ** gamma
-				X, Y, Z = (v / Y * Yy for v in (X, Y, Z))
-			# Apply black offset
-			beta = self.black_cdm2 / self.white_cdm2
-			bp = [E * beta for E in get_whitepoint(self.rgb_space[1])]
-			X, Y, Z = blend_blackpoint(X, Y, Z, (0, 0, 0), bp)
-			return X, Y, Z
+		if v == 1:
+			return 1.0
+		a = 0.17883277
+		b = 1 - 4 * a
+		c = 0.5 - a * math.log(4 * a)
+		if inverse:
+			# Non-linear HLG signal to relative scene linear light
+			if 0 <= v <= 1 / 2.:
+				v = v ** 2 / 3.
+			else:
+				v = (math.exp((v - c) / a) + b) / 12.
 		else:
-			alpha = (self.white_cdm2 - self.black_cdm2)
-			# Rec. 2020 relative luminance coefficients
-			YS = 0.2627 * RGB_S[0] + 0.6780 * RGB_S[1] + 0.0593 * RGB_S[2]
-			return alpha * YS ** gamma + self.black_cdm2
+			# Relative scene linear light to non-linear HLG signal
+			if 0 <= v <= 1 / 12.:
+				v = math.sqrt(3 * v)
+			else:
+				v = a * math.log(12 * v - b) + c
+		return v
+
+	def eotf(self, RGB, inverse=False, apply_black_offset=True):
+		"""
+		Hybrid Log Gamma (HLG) EOTF
+		
+		Non-linear HLG signal to display light, or inverse
+		
+		Input domain 0..1
+		Output range 0..1
+		
+		"""
+		if isinstance(RGB, (float, int)):
+			R, G, B = (RGB,) * 3
+		else:
+			R, G, B = RGB
+		if inverse:
+			# Display light -> relative scene linear light -> HLG signal
+			R, G, B = (self.oetf(v) for v in self.ootf((R, G, B), True,
+													   apply_black_offset))
+		else:
+			# HLG signal -> relative scene linear light -> display light
+			R, G, B = self.ootf([self.oetf(v, True) for v in (R, G, B)], False,
+								apply_black_offset)
+		return G if isinstance(RGB, (float, int)) else (R, G, B)
+		
+
+	def ootf(self, RGB, inverse=False, apply_black_offset=True):
+		"""
+		Hybrid Log Gamma (HLG) OOTF
+
+		Relative scene linear light to display light, or inverse
+		
+		Input domain 0..1
+		Output range 0..1
+		
+		"""
+		if isinstance(RGB, (float, int)):
+			R, G, B = (RGB,) * 3
+		else:
+			R, G, B = RGB
+		if apply_black_offset:
+			black_cdm2 = float(self.black_cdm2)
+		else:
+			black_cdm2 = 0
+		alpha = (self.white_cdm2 - black_cdm2) / self.white_cdm2
+		beta = black_cdm2 / self.white_cdm2
+		Y = 0.2627 * R + 0.6780 * G + 0.0593 * B
+		if inverse:
+			if Y > beta:
+				R, G, B = (((Y - beta) / alpha) ** ((1 - self.gamma) / self.gamma) *
+						   ((v - beta) / alpha) for v in (R, G, B))
+			else:
+				R, G, B = 0, 0, 0
+		else:
+			if Y:
+				Y **= (self.gamma - 1)
+			R, G, B = (alpha * Y * E + beta for E in (R, G, B))
+		return G if isinstance(RGB, (float, int)) else (R, G, B)
+
+	def RGB2XYZ(self, R, G, B, apply_black_offset=True):
+		""" Non-linear HLG signal to display XYZ """
+		X, Y, Z = self.rgb_space[-1] * [self.oetf(v, True) for v in (R, G, B)]
+		X, Y, Z = (max(v, 0) for v in (X, Y, Z))
+		Yy = self.ootf(Y, apply_black_offset=False)
+		if Y:
+			X, Y, Z = (v / Y * Yy for v in (X, Y, Z))
+		else:
+			X, Y, Z = (v * Yy for v in self.rgb_space[1])
+		if apply_black_offset:
+			beta = self.ootf(0)
+			bp_out = [v * beta for v in self.rgb_space[1]]
+			X, Y, Z = apply_bpc(X, Y, Z, (0, 0, 0), bp_out, self.rgb_space[1])
+		return X, Y, Z
+
+	def XYZ2RGB(self, X, Y, Z, apply_black_offset=True):
+		""" Display XYZ to non-linear HLG signal """
+		if apply_black_offset:
+			beta = self.ootf(0)
+			bp_in = [v * beta for v in self.rgb_space[1]]
+			X, Y, Z = apply_bpc(X, Y, Z, bp_in, (0, 0, 0), self.rgb_space[1])
+		Yy = self.ootf(Y, True, apply_black_offset=False)
+		if Y:
+			X, Y, Z = (v / Y * Yy for v in (X, Y, Z))
+		R, G, B = self.rgb_space[-1].inverted() * (X, Y, Z)
+		R, G, B = (max(v, 0) for v in (R, G, B))
+		R, G, B = [self.oetf(v) for v in (R, G, B)]
+		return R, G, B
 
 
 rgb_spaces = {
@@ -268,6 +317,7 @@ rgb_spaces = {
 	"ColorMatch RGB":   (1.8,             "D50",                   (0.6300, 0.3400, 0.274884), (0.2950, 0.6050, 0.658132), (0.1500, 0.0750, 0.066985)),
 #	"DCDM X'Y'Z'":      (2.6,             "E",                     (1.0000, 0.0000, 0.000000), (0.0000, 1.0000, 1.000000), (0.0000, 0.0000, 0.000000)),
 	"DCI P3":           (2.6,             (0.89459, 1.0, 0.95442), (0.6800, 0.3200, 0.209475), (0.2650, 0.6900, 0.721592), (0.1500, 0.0600, 0.068903)),
+	"DCI P3 D65":       (2.6,             "D65",                   (0.6800, 0.3200, 0.209475), (0.2650, 0.6900, 0.721592), (0.1500, 0.0600, 0.068903)),
 	"Don RGB 4":        (2.2,             "D50",                   (0.6960, 0.3000, 0.278350), (0.2150, 0.7650, 0.687970), (0.1300, 0.0350, 0.033680)),
 	"ECI RGB":          (1.8,             "D50",                   (0.6700, 0.3300, 0.320250), (0.2100, 0.7100, 0.602071), (0.1400, 0.0800, 0.077679)),
 	"ECI RGB v2":       (-3.0,            "D50",                   (0.6700, 0.3300, 0.320250), (0.2100, 0.7100, 0.602071), (0.1400, 0.0800, 0.077679)),
@@ -277,7 +327,6 @@ rgb_spaces = {
 	"ProPhoto RGB":     (1.8,             "D50",                   (0.7347, 0.2653, 0.288040), (0.1596, 0.8404, 0.711874), (0.0366, 0.0001, 0.000086)),
 	"Rec. 709":         (-709,            "D65",                   (0.6400, 0.3300, 0.212656), (0.3000, 0.6000, 0.715158), (0.1500, 0.0600, 0.072186)),
 	"Rec. 2020":        (-709,            "D65",                   (0.7080, 0.2920, 0.262694), (0.1700, 0.7970, 0.678009), (0.1310, 0.0460, 0.059297)),
-	"Rec. 2020 ST2084": (-2084,           "D65",                   (0.7080, 0.2920, 0.262694), (0.1700, 0.7970, 0.678009), (0.1310, 0.0460, 0.059297)),
 	"SMPTE-C":          (2.2,             "D65",                   (0.6300, 0.3400, 0.212395), (0.3100, 0.5950, 0.701049), (0.1550, 0.0700, 0.086556)),
 	"SMPTE 240M":       (-240,            "D65",                   (0.6300, 0.3400, 0.212395), (0.3100, 0.5950, 0.701049), (0.1550, 0.0700, 0.086556)),
 	"sRGB":             (-2.4,            "D65",                   (0.6400, 0.3300, 0.212656), (0.3000, 0.6000, 0.715158), (0.1500, 0.0600, 0.072186)),
@@ -1023,8 +1072,8 @@ def Lab2RGB(L, a, b, rgb_space=None, scale=1.0, round_=False, clamp=True,
 	""" Convert from Lab to RGB """
 	X, Y, Z = Lab2XYZ(L, a, b, whitepoint)
 	if not noadapt:
-		X, Y, Z = adapt(X, Y, Z, whitepoint_source, rgb_space[1] if rgb_space
-													else "D65", cat)
+		rgb_space = get_rgb_space(rgb_space)
+		X, Y, Z = adapt(X, Y, Z, whitepoint_source, rgb_space[1], cat)
 	return XYZ2RGB(X, Y, Z, rgb_space, scale, round_, clamp)
 
 
@@ -1146,55 +1195,46 @@ def RGB2HSV(R, G, B, scale=1.0):
 	return tuple(v * scale for v in colorsys.rgb_to_hsv(R, G, B))
 
 
-def RGB2ICtCp(R, G, B, rgb_space="Rec. 2020 ST2084"):
+def LinearRGB2ICtCp(R, G, B, oetf=lambda FD: specialpow(FD, 1.0 / -2084)):
+	""" Rec. 2020 linear RGB to non-linear ICtCp """
 	# http://www.dolby.com/us/en/technologies/dolby-vision/ICtCp-white-paper.pdf
-	rgb2lms_matrix = Matrix3x3([[1688 / 4096., 2146 / 4096., 262 / 4096.],
-								[683 / 4096., 2951 / 4096., 462 / 4096.],
-								[99 / 4096., 309 / 4096., 3688 / 4096.]])
-	LMS = rgb2lms_matrix * (R, G, B)
-	L_, M_, S_ = (specialpow(FD, 1.0 / -2084) for FD in LMS)
-	L_M_S_2ICtCp_matrix = Matrix3x3([[.5, .5, 0],
-									 [6610 / 4096., -13613 / 4096., 7003 / 4096.],
-									 [17933 / 4096., -17390 / 4096., -543 / 4096.]])
+	LMS = LinearRGB2LMS_matrix * (R, G, B)
+	L_, M_, S_ = (oetf(FD) for FD in LMS)
 	I, Ct, Cp = L_M_S_2ICtCp_matrix * (L_, M_, S_)
 	return I, Ct, Cp
 
 
-def ICtCp2RGB(I, Ct, Cp, rgb_space="Rec. 2020 ST2084"):
+def ICtCp2LinearRGB(I, Ct, Cp, eotf=lambda v: specialpow(v, -2084)):
+	""" Non-linear ICtCp to Rec. 2020 linear RGB """
 	# http://www.dolby.com/us/en/technologies/dolby-vision/ICtCp-white-paper.pdf	
-	L_M_S_2ICtCp_matrix = Matrix3x3([[.5, .5, 0],
-									 [6610 / 4096., -13613 / 4096., 7003 / 4096.],
-									 [17933 / 4096., -17390 / 4096., -543 / 4096.]])
-	L_M_S_ = L_M_S_2ICtCp_matrix.inverted() * (I, Ct, Cp)
-	L, M, S = (specialpow(v, -2084) for v in L_M_S_)
-	rgb2lms_matrix = Matrix3x3([[1688 / 4096., 2146 / 4096., 262 / 4096.],
-								[683 / 4096., 2951 / 4096., 462 / 4096.],
-								[99 / 4096., 309 / 4096., 3688 / 4096.]])
-	R, G, B = rgb2lms_matrix.inverted() * (L, M, S)
+	L_M_S_ = ICtCp2L_M_S__matrix * (I, Ct, Cp)
+	L, M, S = (eotf(v) for v in L_M_S_)
+	R, G, B = LMS2LinearRGB_matrix * (L, M, S)
 	return R, G, B
 
 
-def XYZ2ICtCp(X, Y, Z, rgb_space="Rec. 2020 ST2084", clamp=False):
-	R, G, B = (specialpow(v, -2084) for v in XYZ2RGB(X, Y, Z, rgb_space,
-													 clamp=clamp))
-	return RGB2ICtCp(R, G, B, rgb_space)
+def XYZ2ICtCp(X, Y, Z, rgb_space="Rec. 2020", clamp=False,
+			  oetf=lambda E: specialpow(E, 1.0 / -2084)):
+	R, G, B = XYZ2RGB(X, Y, Z, rgb_space, clamp=clamp, oetf=lambda v: v)
+	return LinearRGB2ICtCp(R, G, B, oetf)
 
 
-def ICtCp2XYZ(I, Ct, Cp, rgb_space="Rec. 2020 ST2084"):
-	R, G, B = (specialpow(v, 1.0 / -2084) for v in ICtCp2RGB(I, Ct, Cp, rgb_space))
-	return RGB2XYZ(R, G, B, rgb_space)
+def ICtCp2XYZ(I, Ct, Cp, rgb_space="Rec. 2020",
+			  eotf=lambda v: specialpow(v, -2084)):
+	R, G, B = ICtCp2LinearRGB(I, Ct, Cp, eotf)
+	return RGB2XYZ(R, G, B, rgb_space, eotf=lambda v: v)
 
 
 def RGB2Lab(R, G, B, rgb_space=None, whitepoint=None, noadapt=False,
 			cat="Bradford"):
 	X, Y, Z = RGB2XYZ(R, G, B, rgb_space, scale=100)
 	if not noadapt:
-		X, Y, Z = adapt(X, Y, Z, rgb_space[1] if rgb_space else "D65",
-						whitepoint, cat)
+		rgb_space = get_rgb_space(rgb_space)
+		X, Y, Z = adapt(X, Y, Z, rgb_space[1], whitepoint, cat)
 	return XYZ2Lab(X, Y, Z, whitepoint=whitepoint)
 
 
-def RGB2XYZ(R, G, B, rgb_space=None, scale=1.0):
+def RGB2XYZ(R, G, B, rgb_space=None, scale=1.0, eotf=None):
 	"""
 	Convert from RGB to XYZ.
 	
@@ -1241,7 +1281,9 @@ def RGB2XYZ(R, G, B, rgb_space=None, scale=1.0):
 			gamma = trc[i]
 		else:
 			gamma = trc
-		if isinstance(gamma, (list, tuple)):
+		if eotf:
+			RGB[i] = eotf(v)
+		elif isinstance(gamma, (list, tuple)):
 			RGB[i] = interp(v, [n / float(len(gamma) - 1) for n in
 							    xrange(len(gamma))], gamma)
 		else:
@@ -1250,9 +1292,10 @@ def RGB2XYZ(R, G, B, rgb_space=None, scale=1.0):
 	return tuple(v * scale for v in XYZ)
 
 
-def RGB2YCbCr(R, G, B, rgb_space="NTSC 1953", bits=8):
+def RGB2YCbCr(R, G, B, rgb_space="NTSC 1953", bits=8, fullrange=False):
 	""" R'G'B' to Y'CbCr quantized to n bits """
-	return YPbPr2YCbCr(*RGB2YPbPr(R, G, B, rgb_space), bits=bits)
+	return YPbPr2YCbCr(*RGB2YPbPr(R, G, B, rgb_space), bits=bits,
+					   fullrange=fullrange)
 
 
 def RGB2YPbPr(R, G, B, rgb_space="NTSC 1953"):
@@ -1267,8 +1310,8 @@ def RGB2YPbPr_matrix(rgb_space="NTSC 1953"):
 		ndigits = 3
 	else:
 		ndigits = 4
-	KR = round(rY, ndigits)
-	KB = round(bY, ndigits)
+	KR = round((matrix * (1, 0, 0))[1], ndigits)
+	KB = round((matrix * (0, 0, 1))[1], ndigits)
 	KG = 1.0 - KR - KB
 	Pb_scale = ((1 - KB) / 0.5)
 	Pr_scale = ((1 - KR) / 0.5)
@@ -1277,26 +1320,33 @@ def RGB2YPbPr_matrix(rgb_space="NTSC 1953"):
 					  [0.5, -KG / Pr_scale, -KB / Pr_scale]])
 
 
-def YCbCr2YPbPr(Y, Cb, Cr, bits=8):
+def YCbCr2YPbPr(Y, Cb, Cr, bits=8, fullrange=False):
 	""" Y'CbCr to Y'PbPr """
 	bitlevels = 2 ** bits
-	Yblack = 16 / 256.0 * bitlevels
-	Yscale = 219 / 256.0 * bitlevels
-	Y -= Yblack
+	if not fullrange:
+		Yblack = 16
+		Ywhite = 235
+		Cmax = 240
+	else:
+		Yblack = 0
+		Ywhite = 255
+		Cmax = 255
+	Yscale = (Ywhite - Yblack) / 256.0 * bitlevels
+	Y -= Yblack / 256.0 * bitlevels
 	Y /= Yscale
 	Cneutral = 128 / 256.0 * bitlevels
-	Cscale = 224 / 256.0 * bitlevels
-	Cb -= Cneutral
-	Cb /= Cscale
-	Cr -= Cneutral
-	Cr /= Cscale
-	return Y, Cb, Cr
+	Cscale = (Cmax - Yblack) / 256.0 * bitlevels
+	Pb = Cb - Cneutral
+	Pb /= Cscale
+	Pr = Cr - Cneutral
+	Pr /= Cscale
+	return Y, Pb, Pr
 
 
-def YCbCr2RGB(Y, Cb, Cr, rgb_space="NTSC 1953", bits=8, scale=1.0, round_=False,
-			  clamp=True):
+def YCbCr2RGB(Y, Cb, Cr, rgb_space="NTSC 1953", bits=8, fullrange=False,
+			  scale=1.0, round_=False, clamp=True):
 	""" Y'CbCr to R'G'B' """
-	Y, Pb, Pr = YCbCr2YPbPr(Y, Cb, Cr, bits)
+	Y, Pb, Pr = YCbCr2YPbPr(Y, Cb, Cr, bits, fullrange)
 	return YPbPr2RGB(Y, Pb, Pr, rgb_space, scale, round_, clamp)
 
 
@@ -1313,17 +1363,26 @@ def YPbPr2RGB(Y, Pb, Pr, rgb_space="NTSC 1953", scale=1.0, round_=False,
 	return RGB
 
 
-def YPbPr2YCbCr(Y, Pb, Pr, bits=8):
-	""" Y'PbPr' to Y'CbCr quantized to n bits """
+def YPbPr2YCbCr(Y, Pb, Pr, bits=8, fullrange=False):
+	""" Y'PbPr to Y'CbCr quantized to n bits """
 	bitlevels = 2 ** bits
-	Yblack = 16 / 256.0 * bitlevels
-	Yscale = 219 / 256.0 * bitlevels
-	Y = Yblack + Yscale * Y
+	if not fullrange:
+		Yblack = 16
+		Ywhite = 235
+		Cmax = 240
+	else:
+		Yblack = 0
+		Ywhite = 255
+		Cmax = 255
+	Yscale = (Ywhite - Yblack) / 256.0 * bitlevels
+	Y = Yblack / 256.0 * bitlevels + Yscale * Y
 	Cneutral = 128 / 256.0 * bitlevels
-	Cscale = 224 / 256.0 * bitlevels
+	Cscale = (Cmax - Yblack) / 256.0 * bitlevels
 	Cb = Cneutral + Cscale * Pb
 	Cr = Cneutral + Cscale * Pr
-	Y, Cb, Cr = (int(round(v)) for v in (Y, Cb, Cr))
+	# In fullrange mode, Cb and Cr can reach 255.5, so we need to clamp to 255
+	# Follow ITU-T Rec. T.871 (JPEG)
+	Y, Cb, Cr = (min(max(int(round(v)), 0), 255) for v in (Y, Cb, Cr))
 	return Y, Cb, Cr
 
 
@@ -1368,27 +1427,61 @@ def rgb_to_xyz_matrix(rx, ry, gx, gy, bx, by, whitepoint=None, scale=1.0):
 					  (Sr * Zr, Sg * Zg, Sb * Zb)))
 
 
+def find_primaries_wp_xy_rgb_space_name(xy, rgb_space_names=None,
+												 digits=4):
+	"""
+	Given primaries and whitepoint xy as list, find matching RGB space by
+	comparing primaries and whitepoint (fuzzy match rounded to n digits) and
+	return its name (or None if no match)
+	
+	"""
+	for i, rgb_space_name in enumerate(rgb_space_names or rgb_spaces.iterkeys()):
+		if not rgb_space_names and rgb_space_name in ("ECI RGB", "ECI RGB v2",
+													  "SMPTE 240M", "sRGB"):
+			# Skip in favor of base color space (i.e. NTSC 1953, SMPTE-C and
+			# Rec. 709)
+			continue
+		if get_rgb_space_primaries_wp_xy(rgb_space_name, digits) == xy:
+			return rgb_space_name
+
+
 def get_rgb_space(rgb_space=None, scale=1.0):
 	""" Return gamma, whitepoint, primaries and RGB -> XYZ matrix """
 	if not rgb_space:
 		rgb_space = "sRGB"
 	if isinstance(rgb_space, basestring):
 		rgb_space = rgb_spaces[rgb_space]
-	cachehash = repr(rgb_space[:5]), scale
+	cachehash = tuple(map(id, rgb_space[:5])), scale
 	cache = get_rgb_space.cache.get(cachehash, None)
 	if cache:
 		return cache
 	gamma = rgb_space[0] or rgb_spaces["sRGB"][0]
 	whitepoint = get_whitepoint(rgb_space[1] or rgb_spaces["sRGB"][1], scale)
-	rx, ry, rY = rgb_space[2] or rgb_spaces["sRGB"][2]
-	gx, gy, gY = rgb_space[3] or rgb_spaces["sRGB"][3]
-	bx, by, bY = rgb_space[4] or rgb_spaces["sRGB"][4]
+	rx, ry, rY = rxyY = rgb_space[2] or rgb_spaces["sRGB"][2]
+	gx, gy, gY = gxyY = rgb_space[3] or rgb_spaces["sRGB"][3]
+	bx, by, bY = bxyY = rgb_space[4] or rgb_spaces["sRGB"][4]
 	matrix = rgb_to_xyz_matrix(rx, ry, gx, gy, bx, by, whitepoint, scale)
-	rgb_space = gamma, whitepoint, (rx, ry, rY), (gx, gy, gY), (bx, by, bY), matrix
+	rgb_space = gamma, whitepoint, rxyY, gxyY, bxyY, matrix
 	get_rgb_space.cache[cachehash] = rgb_space
 	return rgb_space
 
 
+def get_rgb_space_primaries_wp_xy(rgb_space=None, digits=4):
+	"""
+	Given RGB space, get primaries and whitepoint xy, optionally rounded to n
+	digits (default 4)
+	
+	"""
+	rgb_space = get_rgb_space(rgb_space)
+	xy = []
+	for i in xrange(3):
+		xy.extend(rgb_space[2:][i][:2])
+	xy.extend(XYZ2xyY(*get_whitepoint(rgb_space[1]))[:2])
+	if digits:
+		xy = [round(v, digits) for v in xy]
+	return xy
+
+
 get_rgb_space.cache = {}
 
 
@@ -1453,8 +1546,8 @@ get_whitepoint.cache = {}
 
 def make_monotonically_increasing(iterable, passes=0, window=None):
 	"""
-	Given an iterable or sequence, make the values monotonically increasing
-	by linear interpolation.
+	Given an iterable or sequence, make the values strictly monotonically
+	increasing (no repeated successive values) by linear interpolation.
 	
 	If iterable is a dict, keep the keys of the original.
 	
@@ -1475,18 +1568,13 @@ def make_monotonically_increasing(iterable, passes=0, window=None):
 		values = smooth_avg(values, passes, window)
 	sequence = zip(keys, values)
 	numvalues = len(sequence)
-	i = 0
-	while len(sequence) > i and i < len(sequence) - 2:
-		# Do not remove the last value if it identical to the second to last
-		while len(sequence) > i + 1 and sequence[i + 1][1] <= sequence[i][1]:
-			# Decreasing or identical
-			sequence.pop(i + 1)
-		i += 1
-	if len(sequence) > 1:
-		# Remove the second to last value if it is identical to the last
-		if sequence[-2][1] >= sequence[-1][1]:
-			# Decreasing or identical
-			sequence.pop(-2)
+	s_new = []
+	y_min = sequence[0][1]
+	while sequence:
+		x, y = sequence.pop()
+		if (not s_new or y < s_new[0][1]) and (y > y_min or not sequence):
+			s_new.insert(0, (x, y))
+	sequence = s_new
 	# Interpolate to original size
 	x_new = [item[0] for item in sequence]
 	y = [item[1] for item in sequence]
@@ -1920,7 +2008,8 @@ def XYZ2Luv(X, Y, Z, whitepoint=None):
 	return L, u, v
 
 
-def XYZ2RGB(X, Y, Z, rgb_space=None, scale=1.0, round_=False, clamp=True):
+def XYZ2RGB(X, Y, Z, rgb_space=None, scale=1.0, round_=False, clamp=True,
+			oetf=None):
 	"""
 	Convert from XYZ to RGB.
 	
@@ -1964,7 +2053,9 @@ def XYZ2RGB(X, Y, Z, rgb_space=None, scale=1.0, round_=False, clamp=True):
 			gamma = trc[i]
 		else:
 			gamma = trc
-		if isinstance(gamma, (list, tuple)):
+		if oetf:
+			RGB[i] = oetf(v)
+		elif isinstance(gamma, (list, tuple)):
 			RGB[i] = interp(v, gamma, [n / float(len(gamma) - 1) for n in
 									   xrange(len(gamma))])
 		else:
@@ -2675,6 +2766,122 @@ class BT1886(object):
 		return out
 
 
+class BT2390(object):
+
+	"""
+	Roll-off for SMPTE 2084 (PQ) according to Report ITU-R BT.2390-2 HDR TV
+	
+	"""
+
+	def __init__(self, black_cdm2, white_cdm2, master_black_cdm2=0,
+				 master_white_cdm2=10000):
+		"""
+		Master black and white level are used to tweak the roll-off and clip.
+		
+		"""
+
+		self.black_cdm2 = black_cdm2
+		self.white_cdm2 = white_cdm2
+		self.master_black_cdm2 = master_black_cdm2
+		self.master_white_cdm2 = master_white_cdm2
+	
+		self.ominv = black_cdm2 / 10000.0  # Lmin
+		self.omini = specialpow(self.ominv, 1.0 / -2084)  # Original minLum
+		self.omaxv = white_cdm2 / 10000.0  # Lmax
+		self.omaxi = specialpow(self.omaxv, 1.0 / -2084)  # Original maxLum
+
+		self.oKS = 1.5 * self.omaxi - 0.5
+
+		# BT.2390-2
+		self.mminv = master_black_cdm2 / 10000.0  # LB
+		self.mmini = specialpow(self.mminv, 1.0 / -2084)
+		self.mmaxv = master_white_cdm2 / 10000.0  # LW
+		self.mmaxi = specialpow(self.mmaxv, 1.0 / -2084)
+		self.mini = (self.omini - self.mmini) / (self.mmaxi - self.mmini)  # Normalized minLum
+		self.minv = specialpow(self.mini, -2084)
+		self.maxi = (self.omaxi - self.mmini) / (self.mmaxi - self.mmini)  # Normalized maxLum
+		self.maxv = specialpow(self.maxi, -2084)
+
+		self.KS = 1.5 * self.maxi - 0.5
+
+	@staticmethod
+	def P(B, KS, maxi, maxci=1.0):
+		T = (B - KS) / (1 - KS)
+		E2 = ((2 * T ** 3 - 3 * T ** 2 + 1) * KS + (T ** 3 - 2 * T ** 2 + T) *
+			  (1 - KS) + (-2 * T ** 3 + 3 * T ** 2) * maxi)
+		if maxci < 1:
+			# Clipping for slightly better target display peak luminance usage
+			s = min(((B - KS) / (maxci - KS)) ** 4, 1.0)
+			E2 = E2 * (1 - s) + maxi * s
+		return E2
+
+	def apply(self, v, KS=None, maxi=None, maxci=1.0, mini=None,
+			  mmaxi=None, mmini=None, bpc=False, normalize=True):
+		"""
+		Apply roll-off (E' in, E' out)
+		maxci if < 1.0 applies alterante clip. SHOULD NOT BE USED FOR PQ.
+		
+		"""
+		if KS is None:
+			KS = self.KS
+		if maxi is None:
+			maxi = self.maxi
+		if mini is None:
+			mini = self.mini
+		if mmaxi is None:
+			mmaxi = self.mmaxi
+		if mmini is None:
+			mmini = self.mmini
+		if normalize and mmini is not None and mmaxi is not None:
+			# Normalize PQ values based on mastering display black/white levels
+			E1 = min(max((v - mmini) / (mmaxi - mmini), 0), 1.0)
+		else:
+			E1 = v
+		# BT.2390-3 suggests P[E1] if KS <= E1 <=1, but this results in
+		# division by zero if KS = 1. The correct way is to check for
+		# KS < E1 <=1
+		if KS < E1 <= 1:
+			E2 = self.P(E1, KS, maxi, maxci)
+		else:
+			E2 = E1
+		# BT.2390-3 suggests 0 <= E2 <= 1, but this results in a discontinuity
+		# if KS < 0 (high LB > Lmin, low Lmax, high LW). To avoid this, check
+		# for E2 <= 1 instead
+		if mini and E2 <= 1:
+			# Apply black level lift
+			minLum = mini
+			maxLum = maxi
+			b = minLum
+			# BT.2390-3 suggests E2 + b * (1 - E2) ** 4, but this clips, if
+			# minLum > 0.25, due to a 'dip' in the function. The solution is to
+			# adjust the exponent according to minLum. For minLum <= 0.25
+			# (< 5.15 cd/m2), this will give the same result as 'pure' BT.2390-3
+			if b >= 0:
+				# Only for positive b i.e. minLum >= LB
+				p = min(1.0 / b, 4)
+			else:
+				# For negative b i.e. minLum < LB
+				p = 4
+			E3 = E2 + b * (1 - E2) ** p
+			# If maxLum < 1, and the input value reaches maxLum, the resulting
+			# output value will be higher than maxLum after applying the black
+			# level lift (note that this is *not* a side-effect of the above 
+			# exponent adjustment). Undo this by re-scaling to the nominal output
+			# range [minLum, maxLum].
+			if maxi < 1:
+				# Only re-scale if maxLum < 1. Note that maxLum can be > 1
+				# if Lmax > LW despite E2 <= 1
+				E3 = convert_range(E3, b, maxi + b * (1 - maxi) ** p, b, maxi)
+		else:
+			E3 = E2
+		if bpc:
+			E3 = convert_range(E3, mini, maxi, 0, maxi)
+		if normalize and mmini is not None and mmaxi is not None:
+			# Invert the normalization of the PQ values
+			E3 = E3 * (mmaxi - mmini) + mmini
+		return max(E3, 0)
+
+
 class Matrix3x3(list):
 	
 	""" Simple 3x3 matrix """
@@ -2863,6 +3070,15 @@ LMS2IPT_matrix = Matrix3x3([[ 0.4000,  0.4000,  0.2000],
 							[ 0.8056,  0.3572, -1.1628]])
 IPT2LMS_matrix = LMS2IPT_matrix.inverted()
 
+LinearRGB2LMS_matrix = Matrix3x3([[1688 / 4096., 2146 / 4096., 262 / 4096.],
+								  [683 / 4096., 2951 / 4096., 462 / 4096.],
+								  [99 / 4096., 309 / 4096., 3688 / 4096.]])
+LMS2LinearRGB_matrix = LinearRGB2LMS_matrix.inverted()
+L_M_S_2ICtCp_matrix = Matrix3x3([[.5, .5, 0],
+								 [6610 / 4096., -13613 / 4096., 7003 / 4096.],
+								 [17933 / 4096., -17390 / 4096., -543 / 4096.]])
+ICtCp2L_M_S__matrix = L_M_S_2ICtCp_matrix.inverted()
+
 # Tweaked LMS to IPT matrix to account for CIE 2012 2deg XYZ to LMS matrix
 # From Argyll/icc/icc.c
 LMS2Lpt_matrix = Matrix3x3([[ 0.6585034777870502,  0.1424555300344579,  0.1990409921784920],
